using System;
using Vim.Math3d;

namespace Quadratic.Carto.Orbital;

/// <summary>
/// A list of keplerian orbital elements, representing the orbit around a celestial body.
/// </summary>
/// <remarks>
/// https://orbital-mechanics.space/classical-orbital-elements/orbital-elements-and-the-state-vector.html
/// </remarks>
public sealed class KeplerianElements(
    double ecc, double sma, double inc,
    double lan, double ape, double ta
)
{
    /// <summary>
    /// Eccentricity of the orbit.
    /// </summary>
    public double ecc = ecc;

    /// <summary>
    /// Semi-major axis of the orbit, in meters.
    /// </summary>
    public double sma = sma;

    /// <summary>
    /// Inclination of the orbit, in radians [0, π).
    /// </summary>
    public double inc = inc;

    /// <summary>
    /// Longitude of the ascending node, in radians [0, 2π).
    /// </summary>
    public double lan = lan;

    /// <summary>
    /// Argument of periapsis, in radians [0, 2π).
    /// </summary>
    public double ape = ape;

    /// <summary>
    /// True anomaly, in radians [0, 2π).
    /// </summary>
    public double ta = ta;

    public double IncDegree => inc * 180 / Math.PI;
    public double LanDegree => lan * 180 / Math.PI;
    public double ApeDegree => ape * 180 / Math.PI;
    public double TaDegree => ta * 180 / Math.PI;

    public double Apoapsis => sma * (1 + ecc);
    public double Periapsis => sma * (1 - ecc);

    public double Period(double mu)
    {
        return 2 * Math.PI * Math.Sqrt(sma * sma * sma / mu);
    }

    public override string ToString()
    {
        return $"KeplerianElements(ecc: {ecc:F6}, sma: {sma:F6}, inc: {inc:F6}, lan: {lan:F6}, ape: {ape:F6}, ta: {ta:F6})";
    }

    /// <summary>
    /// Returns the position and velocity of the body at the given time.
    /// </summary>
    /// <param name="mu">The standard gravitational parameter, mu = G(M+m)</param>
    /// <param name="epoch">The time of the query</param>
    /// <returns></returns>
    // TODO: automatically translated, not verified
    public DStateVector MotionStateAt(double mu, double epoch)
    {
        double n = Math.Sqrt(mu / (sma * sma * sma));
        double M = n * (epoch - 0); // Assuming epoch is time since periapsis

        double EA = ta; // Assuming true anomaly is given as eccentric anomaly
        double MA = EA - ecc * Math.Sin(EA);

        double nu = 2 * Math.Atan(Math.Sqrt((1 + ecc) / (1 - ecc)) * Math.Tan(EA / 2));
        double r = sma * (1 - ecc * Math.Cos(EA));
        double h = Math.Sqrt(mu * sma * (1 - ecc * ecc));

        double Om = lan;
        double w = ape;

        double X = r * (Math.Cos(Om) * Math.Cos(w + nu) - Math.Sin(Om) * Math.Sin(w + nu) * Math.Cos(inc));
        double Z = r * (Math.Sin(Om) * Math.Cos(w + nu) + Math.Cos(Om) * Math.Sin(w + nu) * Math.Cos(inc)); // Swapped Y and Z
        double Y = r * (Math.Sin(inc) * Math.Sin(w + nu)); // Swapped Y and Z

        double p = sma * (1 - ecc * ecc);

        double V_X = (X * h * ecc / (r * p)) * Math.Sin(nu) - (h / r) * (Math.Cos(Om) * Math.Sin(w + nu) + Math.Sin(Om) * Math.Cos(w + nu) * Math.Cos(inc));
        double V_Z = (Z * h * ecc / (r * p)) * Math.Sin(nu) - (h / r) * (Math.Sin(Om) * Math.Sin(w + nu) - Math.Cos(Om) * Math.Cos(w + nu) * Math.Cos(inc)); // Swapped V_Y and V_Z
        double V_Y = (Y * h * ecc / (r * p)) * Math.Sin(nu) + (h / r) * (Math.Cos(w + nu) * Math.Sin(inc)); // Swapped V_Y and V_Z

        return new DStateVector(new DVector3(X, Y, Z), new DVector3(V_X, V_Y, V_Z));
    }

    public static (KeplerianElements, double) FromMotionStateFull(DStateVector state, double mu)
    {
        var r_vec = state.Position;
        var v_vec = state.Velocity;

        var h_vec = r_vec.Cross(v_vec);
        var h = h_vec.Length();

        var r = r_vec.Length();
        var v = v_vec.Length();

        var E = 0.5 * (v * v) - mu / r;
        var a = -mu / (2 * E);
        var e = Math.Sqrt(1 - (h * h) / (a * mu));
        var i = Math.Acos(h_vec.Z / h);
        var omega_LAN = Math.Atan2(h_vec.X, -h_vec.Y);

        var lat = Math.Atan2(r_vec.Z / Math.Sin(i), r_vec.X * Math.Cos(omega_LAN) + r_vec.Y * Math.Sin(omega_LAN));
        var p = a * (1 - e * e);
        var nu = Math.Atan2(Math.Sqrt(p / mu) * DVector3.Dot(r_vec, v_vec), p - r);
        var omega_AP = lat - nu;

        var EA = 2 * Math.Atan(Math.Sqrt((1 - e) / (1 + e)) * Math.Tan(nu / 2));
        var n = Math.Sqrt(mu / (a * a * a));
        var T = -1 / n * (EA - e * Math.Sin(EA));

        var keplerianElements = new KeplerianElements(e, a, i, omega_LAN, omega_AP, nu);
        return (keplerianElements, T);
    }

    public static KeplerianElements FromMotionState(DStateVector state, double mu)
    {
        return FromMotionStateFull(state, mu).Item1;
    }

    public static KeplerianElements EquatorialCircular(double sma, double ta)
    {
        return new KeplerianElements(0, sma, 0, 0, 0, ta);
    }

    public static KeplerianElements Circular(double sma, double inc, double lan, double ta)
    {
        return new KeplerianElements(0, sma, inc, lan, 0, ta);
    }
}

public struct DStateVector(DVector3 position, DVector3 velocity)
{
    public DVector3 Position = position;
    public DVector3 Velocity = velocity;
}
